#include <config.h>
#include <string.h>
#include <stdlib.h>

typedef struct {
	unsigned char* buffer;
	unsigned int buffer_size;
	unsigned int used_bytes;

	unsigned char next_byte;
	unsigned char used_bits;
} bitstream_t;

static void
bitstream_grow(bitstream_t* b)
{
	b->buffer_size <<= 1;
	b->buffer = realloc(b->buffer, b->buffer_size);
	if (!b->buffer)
		abort();
}

static void
bitstream_init(bitstream_t* b)
{
	b->used_bytes = 0;
	b->buffer = NULL;
	b->buffer_size = 2048; /* Grows to 4096 right away. */
	b->next_byte = 0;
	b->used_bits = 0;
	bitstream_grow(b);
}

static void
bitstream_write1(bitstream_t* b, unsigned int bit)
{
	b->next_byte |= bit << (7 - b->used_bits);
	b->used_bits++;
	if (b->used_bits >= 8) {
		b->buffer[b->used_bytes] = b->next_byte;
		b->used_bytes++;
		if (b->used_bytes >= b->buffer_size)
			bitstream_grow(b);
		b->next_byte = 0;
		b->used_bits = 0;
	}
}

static void
bitstream_finish(bitstream_t* b)
{
	while (b->used_bits)
		bitstream_write1(b, 0);
}

/* Can write at most ``unsigned int'' bits. */
static void
bitstream_write(bitstream_t* b, unsigned int bits, unsigned int data)
{
	int i;
	for (i = bits - 1; i >= 0; --i) {
		unsigned int bit = (data >> i) & 1;
		bitstream_write1(b, bit);
	}
}

#define LZSS_DICTSIZE 0x2000
#define LZSS_MIN_MATCH 3
/* Four bits for the length. */
#define LZSS_MAX_MATCH (LZSS_MIN_MATCH + 15)

typedef struct {
	unsigned char dict[LZSS_DICTSIZE];
	unsigned int pos;
	unsigned int zero;
} dict_t;

static unsigned char
dict_read(dict_t* dict, unsigned int offset)
{
	return dict->dict[offset % LZSS_DICTSIZE];
}

static void
dict_write(dict_t* dict, unsigned char ch)
{
	if (dict->pos >= LZSS_DICTSIZE) {
		dict->dict[dict->zero++] = ch;
		dict->zero %= LZSS_DICTSIZE;
	} else {
		dict->dict[dict->pos++] = ch;
	}
}

unsigned char*
th_lz0(const unsigned char* in, unsigned int insize, unsigned int* outsize)
{
	bitstream_t bs;
	const unsigned char* in_ptr = in;

	bitstream_init(&bs);

	while (in_ptr - in < insize) {
		bitstream_write1(&bs, 1);
		bitstream_write(&bs, 8, *in_ptr);
		in_ptr++;
	}

	bitstream_finish(&bs);
	*outsize = bs.used_bytes;
	return bs.buffer;
}

unsigned char*
th_lz(const unsigned char* in, unsigned int insize, unsigned int* outsize)
{
	dict_t dict;
	const unsigned char* in_ptr = in;
	unsigned int tmp;
	unsigned int i, j;
	bitstream_t bs;

	bitstream_init(&bs);
	memset(&dict, 0, sizeof(dict_t));
	dict.pos = 1;

	while (in_ptr - in < insize) {
		unsigned int match_len = LZSS_MIN_MATCH - 1;
		unsigned int match_id = 0;
		for (i = dict.pos - 1; i > 0; --i) {
			tmp = 0;
			for (j = 0; j < LZSS_MAX_MATCH && i + j < dict.pos; ++j) {
				/* XXX: The problem is that the future is read somehow... and
				 * the future is expected to be zero, or something else for the
				 * second loops. */
				if (dict.zero && ((i + j) % LZSS_DICTSIZE >= dict.zero))
					break;
				if (in_ptr + j >= in + insize)
					break;
				if (*(in_ptr + j) != dict_read(&dict, i + j))
					break;
				tmp++;
			}
			if (tmp > match_len) {
				match_len = tmp;
				match_id = i;
				if (match_len == LZSS_MAX_MATCH)
					break;
			}
		}

		if (match_len >= LZSS_MIN_MATCH) {
			bitstream_write1(&bs, 0);
			bitstream_write(&bs, 13, match_id);
			bitstream_write(&bs, 4, match_len - LZSS_MIN_MATCH);
		} else {
			match_len = 1;
			bitstream_write1(&bs, 1);
			bitstream_write(&bs, 8, *in_ptr);
		}

		for (i = 0; i < match_len; ++i) {
			if (in_ptr < in + insize)
				dict_write(&dict, *in_ptr++);
			else
				dict.pos++;
		}
	}

	bitstream_finish(&bs);
	*outsize = bs.used_bytes;
	return bs.buffer;
}
